// Copyright (c) Sleipnir contributors

#pragma once

namespace slp {

/**
 * Marker interface for concepts to determine whether a given scalar or matrix
 * type belongs to Sleipnir.
 */
class SleipnirBase {};

}  // namespace slp
