// Copyright (c) Sleipnir contributors

#include "sleipnir/autodiff/variable_matrix.hpp"

template slp::VariableMatrix<double> slp::solve(
    const slp::VariableMatrix<double>& A, const slp::VariableMatrix<double>& B);
